/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include "m6809.h"
#include "sam.h"
#include "pia.h"
#include "logging.h"
#include "snapshot.h"
#include "tape.h"

/* Not yet implemented: CWAI, SWI, SWI2, SWI3 */

/* Condition Code manipulation macros */

#define CC_E 0x80
#define CC_F 0x40
#define CC_H 0x20
#define CC_I 0x10
#define CC_N 0x08
#define CC_Z 0x04
#define CC_V 0x02
#define CC_C 0x01
#define NXORV		((reg_cc & CC_N)^((reg_cc & CC_V)<<2))

#define CLR_HNZVC	reg_cc &= ~(CC_H|CC_N|CC_Z|CC_V|CC_C)
#define CLR_NZV		reg_cc &= ~(CC_N|CC_Z|CC_V)
#define CLR_NZVC	reg_cc &= ~(CC_N|CC_Z|CC_V|CC_C)
#define CLR_Z		reg_cc &= ~(CC_Z)
#define CLR_NZC		reg_cc &= ~(CC_N|CC_Z|CC_C)
#define CLR_ZC		reg_cc &= ~(CC_Z|CC_C)

#define SET_Z(a)	if (!(a)) reg_cc |= CC_Z;
#define SET_N8(a)	reg_cc |= (a&0x80)>>4
#define SET_N16(a)	reg_cc |= (a&0x8000)>>12
#define SET_H(a,b,r)	reg_cc |= ((a^b^r)&0x10)<<1
#define SET_C8(a)	reg_cc |= (a&0x100)>>8
#define SET_C16(a)	reg_cc |= (a&0x10000)>>16
#define SET_V8(a,b,r)	reg_cc |= ((a^b^r^(r>>1))&0x80)>>6
#define SET_V16(a,b,r)	reg_cc |= ((a^b^r^(r>>1))&0x8000)>>14
#define SET_NZ8(a)		SET_N8(a);SET_Z(a);
#define SET_NZ16(a)		SET_N16(a);SET_Z(a);
#define SET_NZC8(a)		SET_N8(a);SET_Z(a);SET_C8(a);
#define SET_NZVC8(a,b,r)	SET_N8(r);SET_Z(r&0xff);SET_V8(a,b,r);SET_C8(r);
#define SET_NZVC16(a,b,r)	SET_N16(r);SET_Z(r&0xffff);SET_V16(a,b,r);SET_C16(r);

typedef int (*cpu_step)(void);

static int bus_readbyte(void);
static int bus_storebyte(void);
static int bus_readword(void);
static int bus_readword1(void);
static int bus_storeword(void);
static int bus_storeword1(void);

static int addr_direct(void);
static int addr_immediate(void);
static int addr_immediate1(void);
static int addr_extended(void);
static int addr_extended1(void);
static int addr_relative(void);
static int addr_relative1(void);
static int addr_indexed(void);
static int addr_indexed1(void);

static int register_cc(void);
static int register_a(void);
static int register_b(void);
static int register_d(void);
static int register_x(void);
static int register_y(void);
static int register_u(void);
static int register_s(void);

static int op_next(void);
static int op_vma(void);
static int op_undefined(void);
static int op_illegal(void);

static int op_abx(void);
static int op_adc(void);
static int op_add(void);
static int op_addword(void);
static int op_andcc(void);
static int op_and(void);
static int op_asl(void);
static int op_asr(void);
static int op_bcc(void);
static int op_bcs(void);
static int op_beq(void);
static int op_bge(void);
static int op_bgt(void);
static int op_bhi(void);
static int op_bit(void);
static int op_ble(void);
static int op_bls(void);
static int op_blt(void);
static int op_bmi(void);
static int op_bne(void);
static int op_bpl(void);
static int op_bra(void);
static int op_bsr(void);
static int op_bvc(void);
static int op_bvs(void);
static int op_clr(void);
static int op_cmp(void);
static int op_cmpword(void);
static int op_com(void);
static int op_da(void);
static int op_dec(void);
static int op_eor(void);
static int op_exg(void);
static int op_inc(void);
static int op_jmp(void);
static int op_jsr1(void);
static int op_jsr(void);
static int op_ld(void);
static int op_ldword(void);
static int op_lea(void);
static int op_lsr(void);
static int op_mul(void);
static int op_neg(void);
static int op_orcc(void);
static int op_or(void);
static int op_page2(void);
static int op_page3(void);
static int op_pshs(void);
static int op_pshu(void);
static int op_puls(void);
static int op_pulu(void);
static int op_rol(void);
static int op_ror(void);
static int op_rts(void);
static int op_sbc(void);
static int op_sex(void);
static int op_sub(void);
static int op_subword(void);
static int op_sync(void);
static int op_tfr(void);
static int op_tst(void);
static int op_tstword(void);
static int op_vector(void);
static int op_vector1(void);
static int op_rti(void);

static Uint16 sex(Uint8 reg);
static Uint16 sex5(Uint8 reg);

static Uint8 opcode;

/* Physical 6809 lines */
int m6809_vma;

static int irq, firq, nmi;

/* 6809 registers */
Uint8 reg_cc, reg_dp;
Uint16 reg_x, reg_y, reg_u, reg_s, reg_pc;
accumulator_t reg_accum;

/* Points to the next instruction to execute */
cpu_step *next_instruction;

/* byte and word are temporary storage areas for data read from memory */
Uint8 byte;
Uint8 byte1;
Uint16 word;
Uint16 word1;
/* the various _ptrs point to either the temporary store above or a register */
Uint8 *byte_ptr = NULL;
Uint8 *byte1_ptr = NULL;
Uint16 *word_ptr = NULL;
Uint16 *word1_ptr = NULL;
Uint16 interrupt_vector;
/* EAD stores the computed effective address */
Uint16 EAD;
/* RA stores the relative address */
Uint16 RA;
/* Tracks which byte of a word we're stacking */
int stack_step = 0;
/* Store post-byte in indexed addressing mode */
int index_mode;
int index_indirect;
int index_state;
/* Which page is the next instruction to come from? */
int instruction_page = 0;
int current_instruction_page = 0;

/* m6809 core "microcode" */

static cpu_step m6809_ucode_interrupt[] = { op_pshs, op_vector, op_vector1, op_jmp, op_next };

static cpu_step m6809_ucode[0x0100][0x40] = {
	/* 0x00 - 0x0F */

	/* 0x00 NEG direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_neg, bus_storebyte, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x03 COM direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_com, bus_storebyte, op_next },
	/* 0x04 LSR direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_lsr, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x06 ROR direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_ror, bus_storebyte, op_next },
	/* 0x07 ASR direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_asr, bus_storebyte, op_next },
	/* 0x08 ASL direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_asl, bus_storebyte, op_next },
	/* 0x09 ROL direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_rol, bus_storebyte, op_next },
	/* 0x0A DEC direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_dec, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x0C INC direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_inc, bus_storebyte, op_next },
	/* 0x0D TST direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_tst, op_vma, op_next },
	/* 0x0E JMP direct */
	{ addr_direct, op_vma, op_jmp, op_next },
	/* 0x0F CLR direct */
	{ addr_direct, op_vma, bus_readbyte, op_vma, op_clr, bus_storebyte, op_next },

	/* 0x10 - 0x1F */

	/* 0x10 Page 2 */
	{ op_page2, op_next },
	/* 0x11 Page 3 */
	{ op_page3, op_next },
	/* 0x12 NOP inherent */
	{ op_vma, op_next },
	/* 0x13 SYNC inherent */
	{ op_sync, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x16 LBRA relative */
	{ addr_relative, addr_relative1, op_vma, op_vma, op_bra, op_next },
	/* 0x17 LBSR relative */
	{ addr_relative, addr_relative1, op_vma, op_vma, op_bsr, op_vma, op_pshs, op_bra, op_next },
	{ op_illegal, op_next },
	/* 0x19 DAA inherent */
	{ register_a, op_da, op_vma, op_next },
	/* 0x1A ORCC immediate */
	{ addr_immediate, bus_readbyte, register_cc, op_orcc, op_next },
	{ op_illegal, op_next },
	/* 0x1C ANDCC immediate */
	{ addr_immediate, bus_readbyte, register_cc, op_andcc, op_next },
	/* 0x1D SEX inherent */
	{ register_d, register_b, op_sex, op_vma, op_next },
	/* 0x1E EXG immediate */
	{ addr_immediate, bus_readbyte, op_exg, op_vma, op_vma, op_vma, op_vma, op_vma, op_vma, op_next },
	/* 0x1F TFR immediate */
	{ addr_immediate, bus_readbyte, op_tfr, op_vma, op_vma, op_vma, op_vma, op_next },

	/* 0x20 - 0x2f */

	/* 0x20 BRA relative */
	{ addr_relative, op_vma, op_bra, op_next },
	/* 0x21 BRN relative */
	{ addr_relative, op_vma, op_next },
	/* 0x22 BHI relative */
	{ addr_relative, op_vma, op_bhi, op_next },
	/* 0x23 BLS relative */
	{ addr_relative, op_vma, op_bls, op_next },
	/* 0x24 BHS/BCC relative */
	{ addr_relative, op_vma, op_bcc, op_next },
	/* 0x25 BLO/BCS relative */
	{ addr_relative, op_vma, op_bcs, op_next },
	/* 0x26 BNE relative */
	{ addr_relative, op_vma, op_bne, op_next },
	/* 0x27 BEQ relative */
	{ addr_relative, op_vma, op_beq, op_next },
	/* 0x28 BVC relative */
	{ addr_relative, op_vma, op_bvc, op_next },
	/* 0x29 BVS relative */
	{ addr_relative, op_vma, op_bvs, op_next },
	/* 0x2A BPL relative */
	{ addr_relative, op_vma, op_bpl, op_next },
	/* 0x2B BMI relative */
	{ addr_relative, op_vma, op_bmi, op_next },
	/* 0x2C BGE relative */
	{ addr_relative, op_vma, op_bge, op_next },
	/* 0x2D BLT relative */
	{ addr_relative, op_vma, op_blt, op_next },
	/* 0x2E BGT relative */
	{ addr_relative, op_vma, op_bgt, op_next },
	/* 0x2F BLE relative */
	{ addr_relative, op_vma, op_ble, op_next },

	/* 0x30 - 0x3F */

	/* 0x30 LEAX indexed */
	{ addr_indexed, addr_indexed1, register_x, op_lea, op_next },
	/* 0x31 LEAY indexed */
	{ addr_indexed, addr_indexed1, register_y, op_lea, op_next },
	/* 0x32 LEAS indexed */
	{ addr_indexed, addr_indexed1, register_s, op_lea, op_next },
	/* 0x33 LEAU indexed */
	{ addr_indexed, addr_indexed1, register_u, op_lea, op_next },
	/* 0x34 PSHS immediate */
	{ addr_immediate, bus_readbyte, op_vma, op_vma, op_vma, op_pshs, op_next },
	/* 0x35 PULS immediate */
	{ addr_immediate, bus_readbyte, op_vma, op_vma, op_puls, op_vma, op_next },
	/* 0x36 PSHU immediate */
	{ addr_immediate, bus_readbyte, op_vma, op_vma, op_vma, op_pshu, op_next },
	/* 0x37 PULU immediate */
	{ addr_immediate, bus_readbyte, op_vma, op_vma, op_pulu, op_vma, op_next },
	{ op_illegal, op_next },
	/* 0x39 RTS inherent */
	{ op_rts, op_puls, op_vma, op_next },
	/* 0x3A ABX inherent */
	{ register_b, register_x, op_abx, op_vma, op_vma, op_next },
	/* 0x3B RTI inherent */
	{ op_vma, op_rti, op_puls, op_vma, op_next },
	/* 0x3C CWAI inherent */
	{ op_undefined },
	/* 0x3D MUL inherent */
	{ register_a, register_b, register_d, op_mul, op_vma, op_vma, op_vma, op_vma, op_vma, op_vma, op_vma, op_vma, op_vma, op_next },
	{ op_illegal, op_next },
	/* 0x3F SWI inherent */
	{ op_undefined },

	/* 0x40 - 0x4f */

	/* 0x40 NEGA inherent */
	{ register_a, op_neg, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x43 COMA inherent */
	{ register_a, op_com, op_next },
	/* 0x44 LSRA inherent */
	{ register_a, op_lsr, op_next },
	{ op_illegal, op_next },
	/* 0x46 RORA inherent */
	{ register_a, op_ror, op_next },
	/* 0x47 ASRA inherent */
	{ register_a, op_asr, op_next },
	/* 0x48 ASLA inherent */
	{ register_a, op_asl, op_next },
	/* 0x49 ROLA inherent */
	{ register_a, op_rol, op_next },
	/* 0x4A DECA inherent */
	{ register_a, op_dec, op_next },
	{ op_illegal, op_next },
	/* 0x4C INCA inherent */
	{ register_a, op_inc, op_next },
	/* 0x4D TSTA inherent */
	{ register_a, op_tst, op_next },
	{ op_illegal, op_next },
	/* 0x4F CLRA inherent */
	{ register_a, op_clr, op_next },

	/* 0x50 - 0x5f */

	/* 0x50 NEGB inherent */
	{ register_b, op_neg, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x53 COMB inherent */
	{ register_b, op_com, op_next },
	/* 0x54 LSRB inherent */
	{ register_b, op_lsr, op_next },
	{ op_illegal, op_next },
	/* 0x56 RORB inherent */
	{ register_b, op_ror, op_next },
	/* 0x57 ASRB inherent */
	{ register_b, op_asr, op_next },
	/* 0x58 ASLB,LSLB inherent */
	{ register_b, op_asl, op_next },
	/* 0x59 ROLB inherent */
	{ register_b, op_rol, op_next },
	/* 0x5A DECB inherent */
	{ register_b, op_dec, op_next },
	{ op_illegal, op_next },
	/* 0x5C INCB inherent */
	{ register_b, op_inc, op_next },
	/* 0x5D TSTB inherent */
	{ register_b, op_tst, op_next },
	{ op_illegal, op_next },
	/* 0x5F CLRB inherent */
	{ register_b, op_clr, op_next },

	/* 0x60 - 0x6F indexed */

	/* 0x60 NEG indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_neg, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x63 COM indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_com, op_vma, bus_storebyte, op_next },
	/* 0x64 LSR indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_lsr, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x66 ROR indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_ror, op_vma, bus_storebyte, op_next },
	/* 0x67 ASR indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_asr, op_vma, bus_storebyte, op_next },
	/* 0x68 ASL indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_asl, op_vma, bus_storebyte, op_next },
	/* 0x69 ROL indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_rol, op_vma, bus_storebyte, op_next },
	/* 0x6A DEC indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_dec, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x6C INC indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_inc, op_vma, bus_storebyte, op_next },
	/* 0x6D TST indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_tst, op_vma, op_vma, op_next },
	/* 0x6E JMP indexed */
	{ addr_indexed, addr_indexed1, op_jmp, op_next },
	/* 0x6F CLR indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, op_clr, op_vma, bus_storebyte, op_next },

	/* 0x70 - 0x7F */

	/* 0x70 NEG extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_neg, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x73 COM extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_com, op_vma, bus_storebyte, op_next },
	/* 0x74 LSR extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_lsr, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x76 ROR extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_ror, op_vma, bus_storebyte, op_next },
	/* 0x77 ASR extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_asr, op_vma, bus_storebyte, op_next },
	/* 0x78 ASL/LSL extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_asl, op_vma, bus_storebyte, op_next },
	/* 0x79 ROL extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_rol, op_vma, bus_storebyte, op_next },
	/* 0x7A DEC extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_dec, op_vma, bus_storebyte, op_next },
	{ op_illegal, op_next },
	/* 0x7C INC extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_inc, op_vma, bus_storebyte, op_next },
	/* 0x7D TST extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_tst, op_vma, op_vma, op_next },
	/* 0x7E JMP extended */
	{ addr_extended, addr_extended1, op_vma, op_jmp, op_next },
	/* 0x7F CLR extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, op_clr, op_vma, bus_storebyte, op_next },

	/* 0x80 - 0x8F */

	/* 0x80 SUBA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_sub, op_next },
	/* 0x81 CMPA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_cmp, op_next },
	/* 0x82 SBCA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_sbc, op_next },
	/* 0x83 SUBD immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_d, op_subword, op_next },
	/* 0x84 ANDA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_and, op_next },
	/* 0x85 BITA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_bit, op_next },
	/* 0x86 LDA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_ld, op_next },
	{ op_illegal, op_next },
	/* 0x88 EORA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_eor, op_next },
	/* 0x89 ADCA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_adc, op_next },
	/* 0x8A ORA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_or, op_next },
	/* 0x8B ADDA immediate */
	{ addr_immediate, bus_readbyte, register_a, op_add, op_next },
	/* 0x8C CMPX immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_x, op_cmpword, op_vma, op_next },
	/* 0x8D BSR relative */
	{ addr_relative, op_vma, op_bsr, op_vma, op_pshs, op_bra, op_next },
	/* 0x8E LDX immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_x, op_ldword, op_next },
	{ op_illegal, op_next },

	/* 0x90 - 0x9F */

	/* 0x90 SUBA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_sub, op_next },
	/* 0x91 CMPA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_cmp, op_next },
	/* 0x92 SBCA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_sbc, op_next },
	/* 0x93 SUBD direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_d, op_subword, op_next },
	/* 0x94 ANDA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_and, op_next },
	/* 0x95 BITA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_bit, op_next },
	/* 0x96 LDA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_ld, op_next },
	/* 0x97 STA direct */
	{ addr_direct, op_vma, register_a, bus_storebyte, op_tst, op_next },
	/* 0x98 EORA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_eor, op_next },
	/* 0x99 ADCA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_adc, op_next },
	/* 0x9A ORA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_or, op_next },
	/* 0x9B ADDA direct */
	{ addr_direct, bus_readbyte, op_vma, register_a, op_add, op_next },
	/* 0x9C CMPX direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_x, op_cmpword, op_next },
	/* 0x9D JSR direct */
	{ addr_direct, op_vma, op_jsr, op_vma, op_pshs, op_jsr1, op_next },
	/* 0x9E LDX direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_x, op_ldword, op_next },
	/* 0x9F STX direct */
	{ addr_direct, op_vma, register_x, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0xA0 - 0xAF */

	/* 0xA0 SUBA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_sub, op_next },
	/* 0xA1 CMPA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_cmp, op_next },
	/* 0xA2 SBCA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_sbc, op_next },
	/* 0xA3 SUBD indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_d, op_subword, op_vma, op_next },
	/* 0xA4 ANDA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_and, op_next },
	/* 0xA5 BITA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_bit, op_next },
	/* 0xA6 LDA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_ld, op_next },
	/* 0xA7 STA indexed */
	{ addr_indexed, addr_indexed1, register_a, bus_storebyte, op_tst, op_next },
	/* 0xA8 EORA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_eor, op_next },
	/* 0xA9 ADCA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_adc, op_next },
	/* 0xAA ORA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_or, op_next },
	/* 0xAB ADDA indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_a, op_add, op_next },
	/* 0xAC CMPX indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_x, op_cmpword, op_vma, op_next },
	/* 0xAD JSR indexed */
	{ addr_indexed, addr_indexed1, op_jsr, op_vma, op_pshs, op_jsr1, op_next },
	/* 0xAE LDX indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_x, op_ldword, op_next },
	/* 0xAF STX indexed */
	{ addr_indexed, addr_indexed1, register_x, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0xB0 - 0xBF */

	/* 0xB0 SUBA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_sub, op_next },
	/* 0xB1 CMPA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_cmp, op_next },
	/* 0xB2 SBCA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_sbc, op_next },
	/* 0xB3 SUBD extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_d, op_subword, op_next },
	/* 0xB4 ANDA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_and, op_next },
	/* 0xB5 BITA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_bit, op_next },
	/* 0xB6 LDA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_ld, op_next },
	/* 0xB7 STA extended */
	{ addr_extended, addr_extended1, op_vma, register_a, bus_storebyte, op_tst, op_next },
	/* 0xB8 EORA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_eor, op_next },
	/* 0xB9 ADCA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_adc, op_next },
	/* 0xBA ORA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_or, op_next },
	/* 0xBB ADDA extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_a, op_add, op_next },
	/* 0xBC CMPX extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_x, op_cmpword, op_next },
	/* 0xBD JSR extended */
	{ addr_extended, addr_extended1, op_vma, op_jsr, op_vma, op_pshs, op_jsr1, op_next },
	/* 0xBE LDX extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_x, op_ldword, op_next },
	/* 0xBF STX extended */
	{ addr_extended, addr_extended1, op_vma, register_x, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0xC0 - 0xCF */

	/* 0xC0 SUBB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_sub, op_next },
	/* 0xC1 CMPB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_cmp, op_next },
	/* 0xC2 SBCB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_sbc, op_next },
	/* 0xC3 ADDD immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_d, op_addword, op_next },
	/* 0xC4 ANDB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_and, op_next },
	/* 0xC5 BITB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_bit, op_next },
	/* 0xC6 LDB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_ld, op_next },
	{ op_illegal, op_next },
	/* 0xC8 EORB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_eor, op_next },
	/* 0xC9 ADCB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_adc, op_next },
	/* 0xCA ORB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_or, op_next },
	/* 0xCB ADDB immediate */
	{ addr_immediate, bus_readbyte, register_b, op_add, op_next },
	/* 0xCC LDD immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_d, op_ldword, op_next },
	{ op_illegal, op_next },
	/* 0xCE LDU immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_u, op_ldword, op_next },
	{ op_illegal, op_next },

	/* 0xD0 - 0xDF */

	/* 0xD0 SUBB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_sub, op_next },
	/* 0xD1 CMPB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_cmp, op_next },
	/* 0xD2 SBCB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_sbc, op_next },
	/* 0xD3 ADDD direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_d, op_addword, op_vma, op_next },
	/* 0xD4 ANDB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_and, op_next },
	/* 0xD5 BITB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_bit, op_next },
	/* 0xD6 LDB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_ld, op_next },
	/* 0xD7 STB direct */
	{ addr_direct, op_vma, register_b, bus_storebyte, op_tst, op_next },
	/* 0xD8 EORB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_eor, op_next },
	/* 0xD9 ADCB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_adc, op_next },
	/* 0xDA ORB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_or, op_next },
	/* 0xDB ADDB direct */
	{ addr_direct, bus_readbyte, op_vma, register_b, op_add, op_next },
	/* 0xDC LDD direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_d, op_ldword, op_next },
	/* 0xDD STD direct */
	{ addr_direct, op_vma, register_d, bus_storeword, bus_storeword1, op_tstword, op_next },
	/* 0xDE LDU direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_u, op_ldword, op_next },
	/* 0xDF STU direct */
	{ addr_direct, op_vma, register_u, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0xE0 - 0xEF */

	/* 0xE0 SUBB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_sub, op_next },
	/* 0xE1 CMPB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_cmp, op_next },
	/* 0xE2 SBCB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_sbc, op_next },
	/* 0xE3 ADDD indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_d, op_addword, op_vma, op_next },
	/* 0xE4 ANDB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_and, op_next },
	/* 0xE5 BITB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_bit, op_next },
	/* 0xE6 LDB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_ld, op_next },
	/* 0xE7 STB indexed */
	{ addr_indexed, addr_indexed1, register_b, bus_storebyte, op_tst, op_next },
	/* 0xE8 EORB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_eor, op_next },
	/* 0xE9 ADCB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_adc, op_next },
	/* 0xEA ORB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_or, op_next },
	/* 0xEB ADDB indexed */
	{ addr_indexed, addr_indexed1, bus_readbyte, register_b, op_add, op_next },
	/* 0xEC LDD indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_d, op_ldword, op_next },
	/* 0xED STD indexed */
	{ addr_indexed, addr_indexed1, register_d, bus_storeword, bus_storeword1, op_tstword, op_next },
	/* 0xEE LDU indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_u, op_ldword, op_next },
	/* 0xEF STU indexed */
	{ addr_indexed, addr_indexed1, register_u, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0xF0 - 0xFF */

	/* 0xF0 SUBB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_sub, op_next },
	/* 0xF1 CMPB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_cmp, op_next },
	/* 0xF2 SBCB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_sbc, op_next },
	/* 0xF3 ADDD extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_d, op_addword, op_next },
	/* 0xF4 ANDB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_and, op_next },
	/* 0xF5 BITB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_bit, op_next },
	/* 0xF6 LDB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_ld, op_next },
	/* 0xF7 STB extended */
	{ addr_extended, addr_extended1, op_vma, register_b, bus_storebyte, op_tst, op_next },
	/* 0xF8 EORB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_eor, op_next },
	/* 0xF9 ADCB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_adc, op_next },
	/* 0xFA ORB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_or, op_next },
	/* 0xFB ADDB extended */
	{ addr_extended, addr_extended1, op_vma, bus_readbyte, register_b, op_add, op_next },
	/* 0xFC LDD extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_d, op_ldword, op_next },
	/* 0xFD STD extended */
	{ addr_extended, addr_extended1, op_vma, register_d, bus_storeword, bus_storeword1, op_tstword, op_next },
	/* 0xFE LDU extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_u, op_ldword, op_next },
	/* 0xFF STU extended */
	{ addr_extended, addr_extended1, op_vma, register_u, bus_storeword, bus_storeword1, op_tstword, op_next },
};

/* 0x10xx - page2 instructions */
static cpu_step m6809_ucode_page2[0x0100][0x40] = {
	/* 0x1000 - 0x100F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1010 - 0x101F */

	{ op_illegal, op_page2, op_next },
	{ op_illegal, op_page2, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1020 - 0x102F */

	{ op_illegal, op_next },
	/* 0x1021 LBRN relative */
	{ addr_relative, addr_relative1, op_vma, op_next },
	/* 0x1022 LBHI relative */
	{ addr_relative, addr_relative1, op_vma, op_bhi, op_next },
	/* 0x1023 LBLS relative */
	{ addr_relative, addr_relative1, op_vma, op_bls, op_next },
	/* 0x1024 LBHS/LBCC relative */
	{ addr_relative, addr_relative1, op_vma, op_bcc, op_next },
	/* 0x1025 LBCS/LBLO relative */
	{ addr_relative, addr_relative1, op_vma, op_bcs, op_next },
	/* 0x1026 LBNE relative */
	{ addr_relative, addr_relative1, op_vma, op_bne, op_next },
	/* 0x1027 LBEQ relative */
	{ addr_relative, addr_relative1, op_vma, op_beq, op_next },
	/* 0x1028 LBVC relative */
	{ addr_relative, addr_relative1, op_vma, op_bvc, op_next },
	/* 0x1029 LBVS relative */
	{ addr_relative, addr_relative1, op_vma, op_bvs, op_next },
	/* 0x102A LBPL relative */
	{ addr_relative, addr_relative1, op_vma, op_bpl, op_next },
	/* 0x102B LBMI relative */
	{ addr_relative, addr_relative1, op_vma, op_bmi, op_next },
	/* 0x102C LBGE relative */
	{ addr_relative, addr_relative1, op_vma, op_bge, op_next },
	/* 0x102D LBLT relative */
	{ addr_relative, addr_relative1, op_vma, op_blt, op_next },
	/* 0x102E LBGT relative */
	{ addr_relative, addr_relative1, op_vma, op_bgt, op_next },
	/* 0x102F LBLE relative */
	{ addr_relative, addr_relative1, op_vma, op_ble, op_next },

	/* 0x1030 - 0x103F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x103F SWI2 inherent */
	{ op_undefined },

	/* 0x1040 - 0x104F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1050 - 0x105F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1060 - 0x106F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1070 - 0x107F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1080 - 0x108F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x1083 CMPD immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_d, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x108C CMPY immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_y, op_cmpword, op_next },
	{ op_illegal, op_next },
	/* 0x108E LDY immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_y, op_ldword, op_next },
	{ op_illegal, op_next },

	/* 0x1090 - 0x109F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x1093 CMPD direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_d, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x109C CMPY direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_y, op_cmpword, op_next },
	{ op_illegal, op_next },
	/* 0x109E LDY direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_y, op_ldword, op_next },
	/* 0x109F STY direct */
	{ addr_direct, op_vma, register_y, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0x10A0 - 0x10AF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10A3 CMPD indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_d, op_cmpword, op_vma, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10AC CMPY indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_y, op_cmpword, op_vma, op_next },
	{ op_illegal, op_next },
	/* 0x10AE LDY indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_y, op_ldword, op_next },
	/* 0x10AF STY indexed */
	{ addr_indexed, addr_indexed1, register_y, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0x10B0 - 0x10BF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10B3 CMPD extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_d, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10BC CMPY extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_y, op_cmpword, op_next },
	{ op_illegal, op_next },
	/* 0x10BE LDY extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_y, op_ldword, op_next },
	/* 0x10BF STY extended */
	{ addr_extended, addr_extended1, op_vma, register_y, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0x10C0 - 0x10CF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10CE LDS immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, register_s, op_ldword, op_next },
	{ op_illegal, op_next },

	/* 0x10D0 - 0x10DF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10DE LDS direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, register_s, op_ldword, op_next },
	/* 0x10DF STS direct */
	{ addr_direct, op_vma, register_s, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0x10E0 - 0x10EF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10EE LDS indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_s, op_ldword, op_next },
	/* 0x10EF STS indexed */
	{ addr_indexed, addr_indexed1, register_s, bus_storeword, bus_storeword1, op_tstword, op_next },

	/* 0x10F0 - 0x10FF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x10FE LDS extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_s, op_ldword, op_next },
	/* 0x10FF STS extended */
	{ addr_extended, addr_extended1, op_vma, register_s, bus_storeword, bus_storeword1, op_tstword, op_next },
};

/* 0x11xx - page3 instructions */
static cpu_step m6809_ucode_page3[0x0100][0x40] = {
	/* 0x1100 - 0x110F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1110 - 0x111F */

	{ op_illegal, op_page3, op_next },
	{ op_illegal, op_page3, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1120 - 0x112F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1130 - 0x113F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x113F SWI3 inherent */
	{ op_undefined },

	/* 0x1140 - 0x114F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1150 - 0x115F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1160 - 0x116F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1170 - 0x117F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1180 - 0x118F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x1183 CMPU immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_u, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x118C CMPS immediate */
	{ addr_immediate, addr_immediate1, bus_readword, bus_readword1, op_vma, register_s, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x1190 - 0x119F */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x1193 CMPU direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_u, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x119C CMPS direct */
	{ addr_direct, op_vma, bus_readword, bus_readword1, op_vma, register_s, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11A0 - 0x11AF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x11A3 CMPU indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_u, op_cmpword, op_vma, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x11AC CMPS indexed */
	{ addr_indexed, addr_indexed1, bus_readword, bus_readword1, register_s, op_cmpword, op_vma, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11B0 - 0x11BF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x11B3 CMPU extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_u, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	/* 0x11BC CMPS extended */
	{ addr_extended, addr_extended1, op_vma, bus_readword, bus_readword1, register_s, op_cmpword, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11C0 - 0x11CF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11D0 - 0x11DF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11E0 - 0x11EF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

	/* 0x11F0 - 0x11FF */

	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },
	{ op_illegal, op_next },

};


/* Main functions */

void m6809_irq(void) {
	irq = 1;
}
void m6809_firq(void) {
	firq = 1;
}

void m6809_init(void) {
	PIA_0B.intfunc = m6809_irq;
	PIA_1B.intfunc = m6809_firq;
}

void m6809_reset(void) {
	reg_cc = 0xff;
	reg_dp = 0x00;
	reg_accum.reg_d = 0x0000;
	reg_x = reg_y = 0x0000;
	reg_u = reg_s = 0x0000;
	reg_pc = 0xfffe;
	instruction_page = 0;
	next_instruction = m6809_ucode[0x7e];  /* JMP extended */
}

void m6809_cycle(void) {
	cpu_step op = *(next_instruction++);
	int ret;
	m6809_vma = 0;
	while ((ret = op()) != 1) {
		if (ret == -1) {
			/* for things like pshs/sync which need to loop 'til
			 * they're done */
			next_instruction--;
			break;
		} else {
			op = *(next_instruction++);
		}
	}
}

/* Core "microcode" operations */

/* Addressing modes */

int addr_direct(void) {
	EAD = reg_dp << 8 | sam_read_byte(reg_pc++);
	return 1;
}
int addr_immediate(void) {
	EAD = reg_pc++;
	return 0;
}
int addr_immediate1(void) {
	reg_pc++;
	return 0;
}
int addr_extended(void) {
	EAD = sam_read_byte(reg_pc++) << 8;
	return 1;
}
int addr_extended1(void) {
	EAD |= sam_read_byte(reg_pc++);
	return 1;
}
int addr_relative(void) {
	RA = sex(sam_read_byte(reg_pc++));
	return 1;
}
int addr_relative1(void) {
	RA = RA << 8 | sam_read_byte(reg_pc++);
	return 1;
}
int addr_indexed(void) {
	Uint8 index_post;
	index_post = sam_read_byte(reg_pc++);
	switch((index_post >> 5) & 0x03) {
		case 0: word_ptr = &reg_x; break;
		case 1: word_ptr = &reg_y; break;
		case 2: word_ptr = &reg_u; break;
		case 3: word_ptr = &reg_s; break;
	}
	if (index_post & 0x80) {
		index_mode = index_post & 0x0f;
		index_indirect = index_post & 0x10;
		index_state = 1;
	} else {
		EAD = *word_ptr + sex5(index_post);
		next_instruction++;  /* skip further processing */
	}
	return 1;
}
int addr_indexed1(void) {
	switch (index_mode) {
		case 0x00:  /* ,R+ */
			switch (index_state++) {
				case 1:
					EAD = *word_ptr;
					*word_ptr += 1;
					break;
				case 2:
					break;
				case 3:
					index_mode = -1;
					break;
			}
			break;
		case 0x01:  /* ,R++ */
			switch (index_state++) {
				case 1:
					EAD = *word_ptr;
					*word_ptr += 2;
					break;
				case 2:
				case 3:
					break;
				case 4:
					index_mode = -1;
					break;
			}
			break;
		case 0x02:  /* ,-R */
			switch (index_state++) {
				case 1:
					*word_ptr -= 1;
					EAD = *word_ptr;
					break;
				case 2:
					break;
				case 3:
					index_mode = -1;
					break;
			}
			break;
		case 0x03:  /* ,--R */
			switch (index_state++) {
				case 1:
					*word_ptr -= 2;
					EAD = *word_ptr;
					break;
				case 2:
				case 3:
					break;
				case 4:
					index_mode = -1;
					break;
			}
			break;
		case 0x04:  /* ,R */
			EAD = *word_ptr;
			index_mode = -1;
			break;
		case 0x05:  /* B,R */
			switch (index_state++) {
				case 1:
					EAD = *word_ptr + sex(reg_accum.reg_ab.reg_b);
					break;
				case 2:
					index_mode = -1;
					break;
			}
			break;
		case 0x06:  /* A,R */
			switch (index_state++) {
				case 1:
					EAD = *word_ptr + sex(reg_accum.reg_ab.reg_a);
					break;
				case 2:
					index_mode = -1;
					break;
			}
			break;
		case 0x08:  /* 8n,R */
			switch (index_state++) {
				case 1:
					byte = sam_read_byte(reg_pc++);
					EAD = *word_ptr + sex(byte);
					break;
				case 2:
					index_mode = -1;
					break;
			}
			break;
		case 0x09:  /* 16n, R */
			switch (index_state++) {
				case 1:
					word = sam_read_byte(reg_pc++) << 8;
					break;
				case 2:
					word |= sam_read_byte(reg_pc++);
					EAD = *word_ptr + word;
					break;
				case 3:
				case 4:
					break;
				case 5:
					index_mode = -1;
					break;
			}
			break;
		case 0x0b:  /* D,R */
			switch (index_state++) {
				case 1:
					EAD = *word_ptr + reg_accum.reg_d;
					break;
				case 3:
				case 4:
				case 5:
					break;
				case 2:
					index_mode = -1;
					break;
			}
			break;
		case 0x0c:  /* 8n,PCR */
			switch (index_state++) {
				case 1:
					byte = sam_read_byte(reg_pc++);
					EAD = reg_pc + sex(byte);
					break;
				case 2:
					index_mode = -1;
					break;
			}
			break;
		case 0x0d:  /* 16n,PCR */
			switch (index_state++) {
				case 1:
					word = sam_read_byte(reg_pc++) << 8;
					break;
				case 2:
					word |= sam_read_byte(reg_pc++);
					EAD = reg_pc + word;
					break;
				case 3:
					index_mode = -1;
					break;
			}
			break;
		case 0x0e:
			index_mode = -1;
			break;
		case 0x0f:
			switch (index_state++) {
				case 1:
					EAD = sam_read_byte(reg_pc++) << 8;
					break;
				case 2:
					EAD |= sam_read_byte(reg_pc++);
					break;
				case 3:
					index_mode = -1;
					break;
			}
			break;
		case 0x10:  /* indirect fetch */
			switch (index_state++) {
				case 1:
					word = sam_read_byte(EAD) << 8;
					break;
				case 2:
					word |= sam_read_byte(EAD+1);
					EAD = word;
					break;
				case 3:
					index_mode = -1;
					break;
			}
			break;
		default:  /* Unimplemented indexed addressing mode */
			LOG_ERROR1("**UNIMPLEMENTED indexed addressing mode %02x\n", index_mode);
			EAD = 0xffff;
			index_mode = -1;
			break;
	}
	if (index_mode < 0) {
		if (index_indirect) {
			index_indirect = 0;
			index_mode = 0x10;
			index_state = 1;
			return -1;
		}
		return 1;
	}
	return -1;
}

/* Instruction fetch */

static int op_next(void) {
	if (snapshot_state && !instruction_page) {
		switch(snapshot_state) {
			case 1: write_snapshot("snap.sna"); break;
			case 2: read_snapshot("snap.sna"); break;
		}
		snapshot_state = 0;
	}
	byte_ptr = &byte;
	word_ptr = &word;
	if (reg_pc == 0xbde7) {
		next_instruction = m6809_ucode[0x39];
		return 1;
	}
	if (reg_pc == 0xbda5) {
		tape_read_bit();
		next_instruction = m6809_ucode[0x39];
		return 1;
	}
	if (irq && !instruction_page) {
		irq = 0;
		if (!(reg_cc & CC_I)) {
			reg_cc |= CC_E;
			byte = 0xff;  /* set up to push all registers */
			interrupt_vector = 0xfff8;
			next_instruction = m6809_ucode_interrupt;
			return 0;
		}
	}
	if (firq && !instruction_page) {
		firq = 0;
		if (!(reg_cc & CC_F)) {
			reg_cc &= ~(CC_E);
			byte = 0x80;  /* set up to only push PC */
			interrupt_vector = 0xfff6;
			next_instruction = m6809_ucode_interrupt;
			return 0;
		}
	}
	opcode = sam_read_byte(reg_pc++);
	current_instruction_page = instruction_page;
	switch(instruction_page) {
		case 2:
			next_instruction = m6809_ucode_page2[opcode];
			break;
		case 3:
			next_instruction = m6809_ucode_page3[opcode];
			break;
		default:
			next_instruction = m6809_ucode[opcode];
			break;
	}
	instruction_page = 0;
	return 1;
}

/* Data fetch */

static int bus_readbyte(void) {
	byte1_ptr = byte_ptr;
	byte_ptr = &byte;
	byte = sam_read_byte(EAD);
	return 1;
}
static int bus_storebyte(void) {
	sam_store_byte(EAD, *byte_ptr);
	return 1;
}
static int bus_readword(void) {
	word1_ptr = word_ptr;
	word_ptr = &word;
	word = sam_read_byte(EAD) << 8;
	return 1;
}
static int bus_readword1(void) {
	word |= sam_read_byte(EAD + 1);
	return 1;
}
static int bus_storeword(void) {
	sam_store_byte(EAD, *word_ptr >> 8);
	return 1;
}
static int bus_storeword1(void) {
	sam_store_byte(EAD + 1, *word_ptr & 0xff);
	return 1;
}

static int op_vma(void) {
	m6809_vma = 1;
	return 1;
}

static int op_undefined(void) {
	LOG_ERROR2("**UNDEFINED INSTRUCTION %d,%02x\n", current_instruction_page, opcode);
	return op_next();
}

static int op_illegal(void) {
	LOG_WARN2("*ILLEGAL INSTRUCTION %d,%02x\n", current_instruction_page, opcode);
	return 1;
}

/* Set up register destinations */

static int register_cc(void) {
	byte1_ptr = byte_ptr;
	byte_ptr = &reg_cc;
	return 0;
}
static int register_a(void) {
	byte1_ptr = byte_ptr;
	byte_ptr = &reg_accum.reg_ab.reg_a;
	return 0;
}
static int register_b(void) {
	byte1_ptr = byte_ptr;
	byte_ptr = &reg_accum.reg_ab.reg_b;
	return 0;
}
static int register_d(void) {
	word1_ptr = word_ptr;
	word_ptr = &reg_accum.reg_d;
	return 0;
}
static int register_x(void) {
	word1_ptr = word_ptr;
	word_ptr = &reg_x;
	return 0;
}
static int register_y(void) {
	word1_ptr = word_ptr;
	word_ptr = &reg_y;
	return 0;
}
static int register_u(void) {
	word1_ptr = word_ptr;
	word_ptr = &reg_u;
	return 0;
}
static int register_s(void) {
	word1_ptr = word_ptr;
	word_ptr = &reg_s;
	return 0;
}

/* Page 2 and Page 3 instructions */

static int op_page2(void) {
	instruction_page = 2;
	return 1;
}
static int op_page3(void) {
	instruction_page = 3;
	return 1;
}

/* Operations */

static int op_abx(void) {
	*word_ptr += *byte_ptr;
	return 0;
}
static int op_adc(void) {
	int result = *byte_ptr + *byte1_ptr + (reg_cc & CC_C);
	CLR_HNZVC;
	SET_NZVC8(*byte_ptr, *byte1_ptr, result);
	SET_H(*byte_ptr, *byte1_ptr, result);
	*byte_ptr = result;
	return 0;
}
static int op_add(void) {
	int result = *byte_ptr + *byte1_ptr;
	CLR_HNZVC;
	SET_NZVC8(*byte_ptr, *byte1_ptr, result);
	SET_H(*byte_ptr, *byte1_ptr, result);
	*byte_ptr = result;
	return 0;
}
static int op_addword(void) {
	int result = *word_ptr + *word1_ptr;
	CLR_NZVC;
	SET_NZVC16(*word_ptr, *word1_ptr, result);
	*word_ptr = result;
	return 0;
}
static int op_and(void) {
	*byte_ptr &= *byte1_ptr;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	return 0;
}
static int op_andcc(void) {
	*byte_ptr &= *byte1_ptr;
	return 0;
}
static int op_asl(void) {
	int result = *byte_ptr << 1;
	CLR_NZVC;
	SET_NZVC8(*byte_ptr, *byte_ptr, result);
	*byte_ptr = result;
	return 0;
}
static int op_asr(void) {
	CLR_NZC;
	reg_cc |= (*byte_ptr & CC_C);
	*byte_ptr = (*byte_ptr & 0x80) | (*byte_ptr >> 1);
	SET_NZ8(*byte_ptr);
	return 0;
}
static int op_bit(void) {
	int result = *byte_ptr & *byte1_ptr;
	CLR_NZV;
	SET_NZ8(result);
	return 0;
}
static int op_cmp(void) {
	int result = *byte_ptr - *byte1_ptr;
	CLR_NZVC;
	SET_NZVC8(*byte_ptr, *byte1_ptr, result);
	return 0;
}
static int op_cmpword(void) {
	int result = *word_ptr - *word1_ptr;
	CLR_NZVC;
	SET_NZVC16(*word_ptr, *word1_ptr, result);
	return 0;
}
static int op_dec(void) {
	*byte_ptr -= 1;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	if (*byte_ptr == 0x7f)
		reg_cc |= CC_V;
	return 0;
}
static int op_eor(void) {
	*byte_ptr ^= *byte1_ptr;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	return 0;
}

static int op_neg(void) {
	int result = ~(*byte_ptr) + 1;
	CLR_NZVC;
	SET_NZVC8(0, *byte_ptr, result);
	*byte_ptr = result;
	return 0;
}
static int op_com(void) {
	*byte_ptr = ~(*byte_ptr);
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	reg_cc |= CC_C;
	return 0;
}
static int op_da(void) {
	Uint8 adjust = 0;
	Uint8 tmp = *byte_ptr;
	if ((tmp&0x0f) >= 0x0a || reg_cc & CC_H) adjust |= 0x06;
	if (tmp >= 0x90 && (tmp&0x0f) >= 0x0a) adjust |= 0x60;
	if (tmp >= 0xa0 || reg_cc & CC_C) adjust |= 0x60;
	*byte_ptr += adjust;
	CLR_NZV;
	SET_NZC8(*byte_ptr);
	return 0;
}
static int op_lsr(void) {
	CLR_NZC;
	reg_cc |= (*byte_ptr & CC_C);
	*byte_ptr >>= 1;
	SET_Z(*byte_ptr);
	return 0;
}
static int op_mul(void) {
	*word_ptr = *byte_ptr * *byte1_ptr;
	CLR_ZC;
	SET_Z(*word_ptr);
	if (*word_ptr & 0x80)
		reg_cc |= CC_C;
	return 0;
}
static int op_ror(void) {
	int result = (reg_cc & CC_C) << 7;
	CLR_NZC;
	reg_cc |= *byte_ptr & CC_C;
	result |= *byte_ptr >> 1;
	SET_NZ8(result);
	*byte_ptr = result;
	return 0;
}
static int op_rol(void) {
	int tmp,result;
	tmp = *byte_ptr;
	result = (reg_cc & CC_C) | (tmp<<1);
	CLR_NZVC;
	SET_NZVC8(tmp,tmp,result);
	*byte_ptr = result;
	return 0;
}
static int op_inc(void) {
	*byte_ptr += 1;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	if (*byte_ptr == 0x80)
		reg_cc |= CC_V;
	return 0;
}
static int op_tst(void) {
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	return 0;
}
static int op_tstword(void) {
	CLR_NZV;
	SET_NZ16(*word_ptr);
	return 0;
}
static int op_jmp(void) {
	reg_pc = EAD;
	return 0;
}
static int op_clr(void) {
	*byte_ptr = 0;
	CLR_NZVC;
	reg_cc |= CC_Z;
	return 0;
}

static int op_or(void) {
	*byte_ptr |= *byte1_ptr;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	return 0;
}
static int op_orcc(void) {
	*byte_ptr |= *byte1_ptr;
	return 0;
}

static int op_ld(void) {
	*byte_ptr = *byte1_ptr;
	CLR_NZV;
	SET_NZ8(*byte_ptr);
	return 0;
}
static int op_ldword(void) {
	*word_ptr = *word1_ptr;
	CLR_NZV;
	SET_NZ16(*word_ptr);
	return 0;
}
static int op_sub(void) {
	int result = *byte_ptr - *byte1_ptr;
	CLR_NZVC;
	SET_NZVC8(*byte_ptr, *byte1_ptr, result);
	*byte_ptr = result;
	return 0;
}
static int op_subword(void) {
	int result = *word_ptr - *word1_ptr;
	CLR_NZVC;
	SET_NZVC16(*word_ptr, *word1_ptr, result);
	*word_ptr = result;
	return 0;
}
static int op_sbc(void) {
	int result = *byte_ptr - *byte1_ptr - (reg_cc & CC_C);
	CLR_NZVC;
	SET_NZVC8(*byte_ptr, *byte1_ptr, result);
	*byte_ptr = result;
	return 0;
}

/* These make the push/pull ops a lot tidier */

#define PUSHWORD(s,r,b) if (byte & b && !stack_step) { sam_store_byte(--s, r & 0xff); stack_step = 1; return -1; }
#define PUSHWORD1(s,r,b) if (byte & b) { byte &= ~(b); sam_store_byte(--s, r >> 8); stack_step = 0; return -1; }
#define PUSHBYTE(s,r,b) if (byte & b) { byte &= ~(b); sam_store_byte(--s, r); return -1; }
#define PULLWORD(s,r,b) if (byte & b && !stack_step) { r = sam_read_byte(s++) << 8; stack_step = 1; return -1; }
#define PULLWORD1(s,r,b) if (byte & b) { byte &= ~(b); r |= sam_read_byte(s++); stack_step = 0; return -1; }
#define PULLBYTE(s,r,b) if (byte & b) { byte &= ~(b); r = sam_read_byte(s++); return -1; }

static int op_pshu(void) {
	PUSHWORD(reg_u, reg_pc, 0x80);
	PUSHWORD1(reg_u, reg_pc, 0x80);
	PUSHWORD(reg_u, reg_s, 0x40);
	PUSHWORD1(reg_u, reg_s, 0x40);
	PUSHWORD(reg_u, reg_y, 0x20);
	PUSHWORD1(reg_u, reg_y, 0x20);
	PUSHWORD(reg_u, reg_x, 0x10);
	PUSHWORD1(reg_u, reg_x, 0x10);
	PUSHBYTE(reg_u, reg_dp, 0x08);
	PUSHBYTE(reg_u, reg_accum.reg_ab.reg_b, 0x04);
	PUSHBYTE(reg_u, reg_accum.reg_ab.reg_a, 0x02);
	PUSHBYTE(reg_u, reg_cc, 0x01);
	return 0;
}
static int op_pshs(void) {
	PUSHWORD(reg_s, reg_pc, 0x80);
	PUSHWORD1(reg_s, reg_pc, 0x80);
	PUSHWORD(reg_s, reg_u, 0x40);
	PUSHWORD1(reg_s, reg_u, 0x40);
	PUSHWORD(reg_s, reg_y, 0x20);
	PUSHWORD1(reg_s, reg_y, 0x20);
	PUSHWORD(reg_s, reg_x, 0x10);
	PUSHWORD1(reg_s, reg_x, 0x10);
	PUSHBYTE(reg_s, reg_dp, 0x08);
	PUSHBYTE(reg_s, reg_accum.reg_ab.reg_b, 0x04);
	PUSHBYTE(reg_s, reg_accum.reg_ab.reg_a, 0x02);
	PUSHBYTE(reg_s, reg_cc, 0x01);
	return 0;
}
static int op_pulu(void) {
	PULLBYTE(reg_u, reg_cc, 0x01);
	PULLBYTE(reg_u, reg_accum.reg_ab.reg_a, 0x02);
	PULLBYTE(reg_u, reg_accum.reg_ab.reg_b, 0x04);
	PULLBYTE(reg_u, reg_dp, 0x08);
	PULLWORD(reg_u, reg_x, 0x10);
	PULLWORD1(reg_u, reg_x, 0x10);
	PULLWORD(reg_u, reg_y, 0x20);
	PULLWORD1(reg_u, reg_y, 0x20);
	PULLWORD(reg_u, reg_s, 0x40);
	PULLWORD1(reg_u, reg_s, 0x40);
	PULLWORD(reg_u, reg_pc, 0x80);
	PULLWORD1(reg_u, reg_pc, 0x80);
	return 0;
}
static int op_puls(void) {
	PULLBYTE(reg_s, reg_cc, 0x01);
	PULLBYTE(reg_s, reg_accum.reg_ab.reg_a, 0x02);
	PULLBYTE(reg_s, reg_accum.reg_ab.reg_b, 0x04);
	PULLBYTE(reg_s, reg_dp, 0x08);
	PULLWORD(reg_s, reg_x, 0x10);
	PULLWORD1(reg_s, reg_x, 0x10);
	PULLWORD(reg_s, reg_y, 0x20);
	PULLWORD1(reg_s, reg_y, 0x20);
	PULLWORD(reg_s, reg_u, 0x40);
	PULLWORD1(reg_s, reg_u, 0x40);
	PULLWORD(reg_s, reg_pc, 0x80);
	PULLWORD1(reg_s, reg_pc, 0x80);
	return 0;
}

static int op_bsr(void) {
	/* set up for PSHS PC */
	byte = 0x80;
	return 1;
}
static int op_bra(void) {
	reg_pc += RA;
	return 0;
}
static int op_bhi(void) {
	if (!(reg_cc & (CC_Z|CC_C))) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bls(void) {
	if (reg_cc & (CC_Z|CC_C)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bcc(void) {
	if (!(reg_cc & CC_C)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bcs(void) {
	if (reg_cc & CC_C) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bne(void) {
	if (!(reg_cc & CC_Z)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_beq(void) {
	if (reg_cc & CC_Z) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bvc(void) {
	if (!(reg_cc & CC_V)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bvs(void) {
	if (reg_cc & CC_V) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bpl(void) {
	if (!(reg_cc & CC_N)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bmi(void) {
	if (reg_cc & CC_N) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bge(void) {
	if (!NXORV) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_blt(void) {
	if (NXORV) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_bgt(void) {
	if (!(NXORV || reg_cc & CC_Z)) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_ble(void) {
	if (NXORV || reg_cc & CC_Z) {
		reg_pc += RA;
		return (instruction_page == 2);
	}
	return 0;
}
static int op_jsr(void) {
	byte = 0x80;
	return 1;
}
static int op_jsr1(void) {
	reg_pc = EAD;
	return 0;
}
static int op_lea(void) {
	*word_ptr = EAD;
	if (word_ptr == &reg_x || word_ptr == &reg_y) {
		CLR_Z;
		SET_Z(*word_ptr);
	}
	return 0;
}

static int op_rts(void) {
	/* Set up for PULS PC */
	byte = 0x80;
	return 1;
}
static int op_sync(void) {
	if (irq || firq || nmi)
		return 1;
	return -1;
}

static int op_rti(void) {
	reg_cc = sam_read_byte(reg_s++);
	if (reg_cc & CC_E) {
		byte = 0xfe;
	} else {
		byte = 0x80;
	}
	return 1;
}

static int op_sex(void) {
	*word_ptr = sex(*byte_ptr);
	CLR_NZV;
	SET_NZ16(*word_ptr);
	return 0;
}

static int op_tfr(void) {
	Uint8 s_byte = 0xff;
	Uint16 s_word = 0xffff;
	Uint8 *source_byte = &s_byte, *dest_byte = &s_byte;
	Uint16 *source_word = &s_word, *dest_word = &s_word;
	switch (byte & 0xf0) {
		case 0x00: source_word = &reg_accum.reg_d; break;
		case 0x10: source_word = &reg_x; break;
		case 0x20: source_word = &reg_y; break;
		case 0x30: source_word = &reg_u; break;
		case 0x40: source_word = &reg_s; break;
		case 0x50: source_word = &reg_pc; break;
		case 0x80: source_byte = &reg_accum.reg_ab.reg_a; break;
		case 0x90: source_byte = &reg_accum.reg_ab.reg_b; break;
		case 0xa0: source_byte = &reg_cc; break;
		case 0xb0: source_byte = &reg_dp; break;
		default: break;
	}
	s_byte = *source_word & 0xff;
	s_word = 0xff00 | *source_byte;
	switch (byte & 0x0f) {
		case 0x00: dest_word = &reg_accum.reg_d; break;
		case 0x01: dest_word = &reg_x; break;
		case 0x02: dest_word = &reg_y; break;
		case 0x03: dest_word = &reg_u; break;
		case 0x04: dest_word = &reg_s; break;
		case 0x05: dest_word = &reg_pc; break;
		case 0x08: dest_byte = &reg_accum.reg_ab.reg_a; break;
		case 0x09: dest_byte = &reg_accum.reg_ab.reg_b; break;
		case 0x0a: dest_byte = &reg_cc; break;
		case 0x0b: dest_byte = &reg_dp; break;
		default: break;
	}
	*dest_byte = *source_byte;
	*dest_word = *source_word;
	return 0;
}

static int op_exg(void) {
	Uint8 s_byte = 0xff, d_byte = 0xff, tmp_byte;
	Uint16 s_word = 0xffff, d_word = 0xffff, tmp_word;
	Uint8 *source_byte = &s_byte, *dest_byte = &d_byte;
	Uint16 *source_word = &s_word, *dest_word = &d_word;
	switch (byte & 0xf0) {
		case 0x00: source_word = &reg_accum.reg_d; break;
		case 0x10: source_word = &reg_x; break;
		case 0x20: source_word = &reg_y; break;
		case 0x30: source_word = &reg_u; break;
		case 0x40: source_word = &reg_s; break;
		case 0x50: source_word = &reg_pc; break;
		case 0x80: source_byte = &reg_accum.reg_ab.reg_a; break;
		case 0x90: source_byte = &reg_accum.reg_ab.reg_b; break;
		case 0xa0: source_byte = &reg_cc; break;
		case 0xb0: source_byte = &reg_dp; break;
		default: break;
	}
	s_byte = *source_word & 0xff;
	s_word = 0xff00 | *source_byte;
	switch (byte & 0x0f) {
		case 0x00: dest_word = &reg_accum.reg_d; break;
		case 0x01: dest_word = &reg_x; break;
		case 0x02: dest_word = &reg_y; break;
		case 0x03: dest_word = &reg_u; break;
		case 0x04: dest_word = &reg_s; break;
		case 0x05: dest_word = &reg_pc; break;
		case 0x08: dest_byte = &reg_accum.reg_ab.reg_a; break;
		case 0x09: dest_byte = &reg_accum.reg_ab.reg_b; break;
		case 0x0a: dest_byte = &reg_cc; break;
		case 0x0b: dest_byte = &reg_dp; break;
		default: break;
	}
	d_byte = *dest_word & 0xff;
	d_word = 0xff00 | *dest_byte;
	tmp_byte = *dest_byte;
	tmp_word = *dest_word;
	*dest_byte = *source_byte;
	*dest_word = *source_word;
	*source_byte = tmp_byte;
	*source_word = tmp_word;
	return 0;
}

static int op_vector(void) {
	EAD = sam_read_byte(interrupt_vector) << 8;
	return 1;
}
static int op_vector1(void) {
	EAD |= sam_read_byte(interrupt_vector + 1);
	return 1;
}

static Uint16 sex(Uint8 reg) {
	return ((reg&0x80)?0xff00:0x0000)|reg;
}
static Uint16 sex5(Uint8 reg) {
	return ((reg&0x10)?0xfff0:0x0000)|(reg&0x0f);
}
