/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <SDL/SDL.h>

#include "sam.h"
#include "pia.h"
#include "logging.h"

static Uint8 *address_decode(Uint16 addr);
static void update_video_mode(void);

extern int trace;

Uint8 ram0[32768];
Uint8 ram1[32768];
Uint8 rom0[32768];
Uint8 rom1[32768];

static int video_mode;
static Uint16 video_base_addr;
static Uint8 *video_ptr;
static int video_xclock, video_yclock;
static int video_xdivisor, video_ydivisor;
static Uint16 video_offset, video_addr, video_line_addr, video_hmask;

static int sam_maptype = 0;
static int sam_page1 = 0;
static int sam_s = 0;

void sam_init(void) {
	memset(ram0, 0, sizeof(ram0));
	memset(ram1, 0, sizeof(ram1));
	memset(rom0, 0, sizeof(rom0));
	memset(rom1, 0, sizeof(rom0));
	if (!load_image("dragon.rom", rom0, sizeof(rom0))) {
		exit(1);
	}
	/* load_image("cart.rom", &rom0[0x4000], 0x3f00); */
}

void sam_reset(void) {
	sam_maptype = 0;
	sam_page1 = 0;
	sam_s = 0;
	video_base_addr = 0;
	video_mode = 0;
	update_video_mode();
	sam_video_vsync();  /* resets video pointers */
}

static Uint8 *address_decode(Uint16 addr) {
	if (addr < 0x8000) {
		sam_s = 0;
		if (sam_page1 == 0) {
			return &ram0[addr];
		}
		return &ram1[addr];
	}
	if (addr < 0xc000 && sam_maptype == 0) {
		sam_s = 1;
		return &rom0[addr-0x8000];
	}
	if (addr < 0xff00) {
		if (sam_maptype) {
			sam_s = 0;
			return &ram1[addr-0x8000];
		}
		sam_s = 1;
		return &rom1[addr-0xc000];
	}
	if (addr < 0xff20) {
		sam_s = 4;
		return NULL;
	}
	if (addr < 0xff40) {
		sam_s = 5;
		return NULL;
	}
	if (addr < 0xff60) {
		sam_s = 6;
		return NULL;
	}
	if (addr >= 0xffc0 && addr < 0xffe0) {
		sam_s = 2;
		return NULL;
	}
	/* $FFF2 - $FFFF jump table, translated to $BFF2 - $BFFF */
	if (addr >= 0xfff2) {
		sam_s = 1;
		return &rom0[addr-0xc000];
	}
	sam_s = 1;
	return (rom0);
}

Uint8 sam_read_byte(Uint16 addr) {
	Uint8 *a = address_decode(addr);

	switch (sam_s) {
		case 4: if ((addr & 3) == 0) return pia_data_read(&PIA_0A);
			if ((addr & 3) == 1) return pia_control_read(&PIA_0A);
			if ((addr & 3) == 2) return pia_data_read(&PIA_0B);
			return pia_control_read(&PIA_0B);
		case 5: if ((addr & 3) == 0) return pia_data_read(&PIA_1A);
			if ((addr & 3) == 1) return pia_control_read(&PIA_1A);
			if ((addr & 3) == 2) return pia_data_read(&PIA_1B);
			return pia_control_read(&PIA_1B);
		case 0:
		case 1:
			return (*a);
		case 2:
		default:
			return 0xff;
	}
}

Uint16 sam_read_word(Uint16 addr) {
	return (sam_read_byte(addr)<<8 | sam_read_byte(addr+1));
}

void sam_store_byte(Uint16 addr, Uint8 byte) {
	Uint8 *a = address_decode(addr);
	switch (sam_s) {
		case 4: if ((addr & 3) == 0) pia_data_write(&PIA_0A, byte);
			if ((addr & 3) == 1) pia_control_write(&PIA_0A, byte);
			if ((addr & 3) == 2) pia_data_write(&PIA_0B, byte);
			if ((addr & 3) == 3) pia_control_write(&PIA_0B, byte);
			return;
		case 5: if ((addr & 3) == 0) pia_data_write(&PIA_1A, byte);
			if ((addr & 3) == 1) pia_control_write(&PIA_1A, byte);
			if ((addr & 3) == 2) pia_data_write(&PIA_1B, byte);
			if ((addr & 3) == 3) pia_control_write(&PIA_1B, byte);
			return;
		case 2:
			if (addr < 0xffc6) {
				/* set video_mode */
				addr -= 0xffc0;
				if (addr & 1) {
					video_mode |= 1 << (addr>>1);
				} else {
					video_mode &= ~(1 << (addr>>1));
				}
				update_video_mode();
				return;
			}
			if (addr < 0xffd4) {
				/* set video_offset */
				addr -= 0xffc6;
				if (addr & 1) {
					video_offset |= 512 << (addr>>1);
				} else {
					video_offset &= ~(512 << (addr>>1));
				}
				return;
			}
			if (addr == 0xffde) sam_maptype = 0;
			if (addr == 0xffdf) sam_maptype = 1;
			return;
		case 0:
			*a = byte;
			return;
		default: break;
	}
}

Uint8 sam_video_read(void) {
	Uint8 ret = *(video_ptr++);
	sam_video_addrskip(1);
	return ret;
}

void sam_video_addrskip(int skip) {
	/* Don't skip if we're in 'DMA' mode */
	if (video_mode != 0x07 || skip != 0x10)
		video_addr += skip;
	if ((video_addr & 0x000f) == 0) {
		if (++video_xclock < video_xdivisor) {
			video_addr -= 0x10;
		} else {
			video_xclock = 0;
		}
		video_ptr = address_decode(video_offset + video_addr);
	}
}

void sam_video_hsync(void) {
	video_addr &= video_hmask;
	if (++video_yclock < video_ydivisor) {
		video_addr = video_line_addr;
	} else {
		video_line_addr = video_addr;
		video_yclock = 0;
	}
	video_ptr = address_decode(video_offset + video_addr);
}

void sam_video_vsync(void) {
	video_xclock = video_yclock = 0;
	video_line_addr = video_addr = 0;
	video_offset &= 0xfe00;
	video_ptr = address_decode(video_offset + video_addr);
}

static void update_video_mode(void) {
	switch (video_mode) {
		default:
		case 0: video_xdivisor = 1;
			video_ydivisor = 12;
			video_hmask = 0xffe0;
			break;
		case 1: video_xdivisor = 3;
			video_ydivisor = 1;
			video_hmask = 0xfff0;
			break;
		case 2: video_xdivisor = 1;
			video_ydivisor = 3;
			video_hmask = 0xffe0;
			break;
		case 3: video_xdivisor = 2;
			video_ydivisor = 1;
			video_hmask = 0xfff0;
			break;
		case 4: video_xdivisor = 1;
			video_ydivisor = 2;
			video_hmask = 0xffe0;
			break;
		case 5: video_xdivisor = 1;
			video_ydivisor = 1;
			video_hmask = 0xfff0;
			break;
		case 6: video_xdivisor = 1;
			video_ydivisor = 1;
			video_hmask = 0xffe0;
			break;
		case 7: video_xdivisor = 1;
			video_ydivisor = 1;
			video_hmask = 0xffff;
			break;
	}
}


int load_image(char *filename, Uint8 *buffer, int size) {
	int fd, r;
	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		LOG_DEBUG2("Failed to read '%s': %s.\n", filename,
				strerror(errno));
		return 0;
	}
	do {
		r = read(fd, (void *)buffer, size);
		if (r == -1) {
			LOG_DEBUG2("Failed to read '%s': %s.\n", filename,
					strerror(errno));
			return 0;
		}
		size -= r;
		buffer += r;
	} while (r && size > 0);
	return 1;
}
