/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>

#include "m6809.h"
#include "sam.h"
#include "vdg.h"
#include "pia.h"
#include "logging.h"

#define STATE_LBORDER 0
#define STATE_DISPLAY 1
#define STATE_RBORDER 2
#define STATE_HBLANK 3
#define STATE_VBLANK 4

#define CYCLES_FRAME 17898
#define CYCLES_LBORDER 4
#define CYCLES_DISPLAY 32
#define CYCLES_RBORDER 4
#define CYCLES_HBLANK 17
#define CYCLES_TBORDER 40
#define CYCLES_BBORDER 40
#define CYCLES_VBLANK 4218

#define LINES_TOP 24
#define LINES_DISPLAY 192
#define LINES_BOTTOM 24

void (*render_func)(Uint8 byte);
static void render_alpha(Uint8 byte);
static void render_g6r(Uint8 byte);
static void render_g6c(Uint8 byte);
static void render_g3r(Uint8 byte);
static void render_g3c(Uint8 byte);
static void render_border(void);
static void vdg_set_mode(Uint8 mode);

static int vdg_gmode;
static int vdg_css;
static int vdg_isgraphics;
static int state;
static int state_count;
static int line;
static Uint8 da0;
static Uint8 da0mask;
static Uint8 fetched_byte;

extern Uint8 vdg_alpha[768];
extern Uint8 vdg_semig[192];

SDL_Surface *screen;
Uint8 *pixel;
Uint8 darkgreen, black;
Uint8 border_colour;
Uint8 bg_colour;
Uint8 fg_colour;
Uint8 vdg_colour[8];
Uint8 *cg_colours;

void vdg_init(void) {
	screen = SDL_SetVideoMode(320, 240, 8, SDL_SWSURFACE);
	if ( screen == NULL ) {
		LOG_ERROR1("Unable to set 320x240x8 video: %s\n",
				SDL_GetError());
		exit(1);
	}
	vdg_colour[0] = SDL_MapRGB(screen->format, 0x00, 0xff, 0x00);
	vdg_colour[1] = SDL_MapRGB(screen->format, 0xff, 0xff, 0x00);
	vdg_colour[2] = SDL_MapRGB(screen->format, 0x00, 0x00, 0xff);
	vdg_colour[3] = SDL_MapRGB(screen->format, 0xff, 0x00, 0x00);
	vdg_colour[4] = SDL_MapRGB(screen->format, 0xff, 0xe0, 0xe0);
	vdg_colour[5] = SDL_MapRGB(screen->format, 0x00, 0xff, 0xff);
	vdg_colour[6] = SDL_MapRGB(screen->format, 0xff, 0x00, 0xff);
	vdg_colour[7] = SDL_MapRGB(screen->format, 0xff, 0xa5, 0x00);
	black = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);
	darkgreen = SDL_MapRGB(screen->format, 0x00, 0x20, 0x00);
	PIA_1B.wrfunc[0] = vdg_set_mode;
	vdg_set_mode(0x00);
}

void vdg_reset(void) {
	state = STATE_RBORDER;
	state_count = CYCLES_TBORDER;
	border_colour = black;
	pixel = (Uint8 *)screen->pixels;
	line = 0;
}

void vdg_cycle(void) {
	switch(state) {
		case STATE_LBORDER:
			render_border();
			if (!(--state_count)) {
				state = STATE_DISPLAY;
				state_count = CYCLES_DISPLAY;
			}
			return;
		case STATE_DISPLAY:
			if ((da0 & da0mask) == 0)
				fetched_byte = sam_video_read();
			render_func(fetched_byte);
			da0 = (da0+1)&da0mask;
			if (!(--state_count)) {
				state = STATE_RBORDER;
				state_count = CYCLES_RBORDER;
				if (!vdg_isgraphics)
					sam_video_addrskip(16);
				sam_video_addrskip(10);
				sam_video_hsync();
			}
			return;
		case STATE_RBORDER:
			render_border();
			if (!(--state_count)) {
				state = STATE_HBLANK;
				state_count = CYCLES_HBLANK;
			}
			return;
		case STATE_HBLANK:
			if (!(--state_count)) {
				line++;
				if (line < LINES_TOP) {
					state = STATE_RBORDER;
					state_count = CYCLES_TBORDER;
					return;
				}
				if (line < LINES_TOP + LINES_DISPLAY) {
					state = STATE_LBORDER;
					state_count = CYCLES_LBORDER;
					return;
				}
				if (line < LINES_TOP + LINES_DISPLAY + LINES_BOTTOM) {
					state = STATE_RBORDER;
					state_count = CYCLES_BBORDER;
					return;
				}
				state = STATE_VBLANK;
				state_count = CYCLES_VBLANK;
				SDL_UpdateRect(screen, 0,0,320,240);
				/* This generates an IRQ */
				pia_set_ca1(&PIA_0B);
				sam_video_vsync();
			}
			return;
		case STATE_VBLANK:
			if (!(--state_count)) {
				state = STATE_RBORDER;
				state_count = CYCLES_TBORDER;
				line = 0;
				pixel = (Uint8 *)screen->pixels;
			}
			return;
	}
}

static void render_alpha(Uint8 byte) {
	Uint8 tmp;
	if (byte & 0x80) {
		tmp = vdg_colour[(byte & 0x70)>>4];
		if ((line % 12) >= 6)
			byte >>= 2;
		*(pixel++) = (byte & 0x02) ? tmp : black;
		*(pixel++) = (byte & 0x02) ? tmp : black;
		*(pixel++) = (byte & 0x02) ? tmp : black;
		*(pixel++) = (byte & 0x02) ? tmp : black;
		*(pixel++) = (byte & 0x01) ? tmp : black;
		*(pixel++) = (byte & 0x01) ? tmp : black;
		*(pixel++) = (byte & 0x01) ? tmp : black;
		*(pixel++) = (byte & 0x01) ? tmp : black;
	} else {
		tmp = vdg_alpha[(byte&0x3f)*12 + line%12];
		if (byte & 0x40)
			tmp = ~tmp;
		*(pixel++) = (tmp & 0x80) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x40) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x20) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x10) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x08) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x04) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x02) ? fg_colour : bg_colour;
		*(pixel++) = (tmp & 0x01) ? fg_colour : bg_colour;
	}
}

static void render_g3c(Uint8 byte) {
	if (da0 & 1)
		byte <<= 4;
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
}

static void render_g3r(Uint8 byte) {
	if (da0 & 1)
		byte <<= 4;
	*(pixel++) = (byte & 0x80) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x80) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x40) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x40) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x20) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x20) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x10) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x10) ? fg_colour : bg_colour;
}

static void render_g6c(Uint8 byte) {
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0xc0) >> 6];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
	*(pixel++) = cg_colours[(byte & 0x30) >> 4];
	*(pixel++) = cg_colours[(byte & 0x0c) >> 2];
	*(pixel++) = cg_colours[(byte & 0x0c) >> 2];
	*(pixel++) = cg_colours[byte & 0x03];
	*(pixel++) = cg_colours[byte & 0x03];
}

static void render_g6r(Uint8 byte) {
	*(pixel++) = (byte & 0x80) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x40) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x20) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x10) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x08) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x04) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x02) ? fg_colour : bg_colour;
	*(pixel++) = (byte & 0x01) ? fg_colour : bg_colour;
}

static void render_border(void) {
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
	*(pixel++) = border_colour;
}

static void vdg_set_mode(Uint8 mode) {
	vdg_css = (mode & 0x08) >> 1;
	vdg_gmode = (mode & 0x70) >> 4;
	vdg_isgraphics = mode & 0x80;
	border_colour = black;
	if (vdg_isgraphics) {
		border_colour = vdg_colour[vdg_css];
		cg_colours = &vdg_colour[vdg_css];
		bg_colour = black;
		fg_colour = border_colour;
		switch (vdg_gmode) {
			case 0: da0mask = 0x01; render_func = render_g3c; break;
			case 1:
			case 3:
			case 5: da0mask = 0x01; render_func = render_g3r; break;
			case 2:
			case 4:
			case 6: da0mask = 0x00; render_func = render_g6c; break;
			default:
			case 7: da0mask = 0x00; render_func = render_g6r; break;
		}
	} else {
		da0mask = 0x00;
		border_colour = black;
		bg_colour = darkgreen;
		if (vdg_css)
			fg_colour = vdg_colour[7];
		else
			fg_colour = vdg_colour[0];
		render_func = render_alpha;
	}
}
